# Deb-o-Matic
#
# Copyright (C) 2015-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from ast import literal_eval
from configparser import NoOptionError
from logging import error

from Debomatic import dom
from .exceptions import DebomaticConffileError


core = {'debomatic':
        {'builduser': str, 'debootstrap': str, 'incoming': str,
         'architecture': str, 'threads': int, 'inotify': bool, 'sleep': int,
         'interval': int, 'logfile': str, 'loglevel': str},
        'distributions': {'list': str, 'blacklist': str, 'mapper': dict},
        'chroots': {'profile': str, 'commands': str},
        'gpg': {'gpg': bool, 'keyring': str},
        'modules': {'modules': bool, 'path': str,
                    'threads': int, 'blacklist': str}}
optional = {'crossbuild': {'crossbuild': bool, 'hostarchitecture': str},
            'dpr': {'dpr': bool, 'prefix': str, 'repository': str}}
modules = {'autopkgtest': {'options': str},
           'blhc': {'options': str},
           'buildcleaner': {'testbuild': bool},
           'lintian': {'options': str},
           'mailer': {'sender': str, 'server': str, 'port': int, 'tls': bool,
                      'authrequired': bool, 'user': str, 'passwd': str,
                      'success': str, 'failure': str, 'lintian': bool},
           'piuparts': {'options': str},
           'repository': {'gpgkey': str, 'pubring': str}}
dists = {'suite': str, 'mirror': str, 'components': str,
         '_extramirrors': str, '_extrapackages': str}


class Parser:

    def __init__(self):
        pass

    def _validate(self, option, section, configtype, element, conffile):
        if not element.has_option(section, option):
            if not option.startswith('_'):
                error(_('Set "%(option)s" in section "%(section)s" '
                        'in %(conffile)s') % {'option': option.strip('_'),
                                              'section': section,
                                              'conffile': conffile})
                raise DebomaticConffileError
        fubar = False
        try:
            if configtype == int:
                _option = element.getint(section, option)
            elif configtype == bool:
                _option = element.getboolean(section, option)
            elif configtype == dict:
                _option = literal_eval(element.get(section, option))
            else:
                _option = element.get(section, option)
            assert(isinstance(_option, configtype))
        except NoOptionError:
            if not option.startswith('_'):
                fubar = True
        except (AssertionError, ValueError):
            fubar = True
        finally:
            if fubar:
                error(_('Option "%(option)s" in section "%(section)s" '
                        'must be %(type)s' % {'option': option,
                                              'section': section,
                                              'type': configtype.__name__}))
                raise DebomaticConffileError

    def parse_configfiles(self):
        if not self.conffile:
            error(_('Configuration file has not been specified'))
            raise DebomaticConffileError
        if not os.path.isfile(self.conffile):
            error(_('Configuration file %s does not exist') % self.conffile)
            raise DebomaticConffileError
        dom.opts.read(self.conffile)
        for section in core:
            if not dom.opts.has_section(section):
                error(_('Section "%(section)s" missing in %(conffile)s') %
                      {'section': section, 'conffile': self.conffile})
                raise DebomaticConffileError
            for option in core[section]:
                self._validate(option, section, core[section][option],
                               dom.opts, self.conffile)
        for section in optional:
            if dom.opts.has_section(section):
                for option in optional[section]:
                    self._validate(option, section, optional[section][option],
                                   dom.opts, self.conffile)
        for section in modules:
            if dom.opts.has_section(section):
                for option in modules[section]:
                    self._validate(option, section, modules[section][option],
                                   dom.opts, self.conffile)
        distfile = dom.opts.get('distributions', 'list')
        if not os.path.isfile(distfile):
            error(_('Distribution file %s does not exist') % distfile)
            raise DebomaticConffileError
        dom.dists.read(distfile)
        for dist in dom.dists.sections():
            for option in dists:
                self._validate(option, dist, dists[option],
                               dom.dists, distfile)



# Deb-o-Matic - RemoveChroots module
#
# Copyright (C) 2018-2022 Luca Falavigna
#
# Authors: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Remove chroots created long ago

import os
from glob import glob
from shutil import rmtree
from subprocess import check_output as call
from time import time


class DebomaticModule_RemoveChroots:

    def __init__(self):
        pass

    def __purge_chroot(self, distribution, directory):
        architecture = directory.split('-')[1]
        chroots = call(['/usr/bin/schroot', '--all-chroots', '-l'])
        sessions = call(['/usr/bin/schroot', '--all-sessions', '-l'])
        chroot = '{0}-{1}-debomatic'.format(distribution, architecture)
        if chroot in chroots.decode('utf-8'):
            if chroot not in sessions.decode('utf-8'):
                for directory in ('/etc/schroot/chroot.d',
                                  '/etc/sbuild/chroot'):
                    for pattern in ('{0}*'.format(chroot),
                                    '*-{0}-{1}-debomatic*'.format(
                                    architecture, distribution)):
                        for dir in glob(os.path.join(directory, pattern)):
                            if os.path.islink(dir):
                                rmtree(os.readlink(dir))
                            os.unlink(dir)

    def periodic(self, args):
        ctime = time()
        if args.opts.has_section('removechroots'):
            delta = args.opts.getint('removechroots', 'days') * 24 * 60 * 60
            for distribution in os.listdir(args.directory):
                chroot = os.path.join(args.directory,
                                      distribution, distribution)
                if os.path.isdir(chroot):
                    ptime = os.stat(chroot).st_mtime
                    if ptime + delta < ctime:
                        self.__purge_chroot(distribution, args.directory)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Deb-o-Matic documentation build configuration file


extensions = []
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'Deb-o-Matic'
copyright = '2007-2022, Luca Falavigna'
version = '0.26'
release = '0.26'
exclude_patterns = ['_build']
pygments_style = 'sphinx'
html_theme = 'default'
html_static_path = ['_static']
html_use_index = True
htmlhelp_basename = 'Deb-o-Maticdoc'
latex_documents = [
    ('index', 'Deb-o-Matic.tex', 'Deb-o-Matic Documentation',
     'Luca Falavigna', 'manual', 'True')]
latex_elements = {
    'classoptions': ',oneside',
    'babel': '\\usepackage[english]{babel}'}
man_pages = [
    ('index', 'deb-o-matic', 'Deb-o-Matic Documentation',
     ['Luca Falavigna'], 1)]



# Deb-o-Matic - UpdateChroots module
#
# Copyright (C) 2018-2022 Luca Falavigna
#
# Authors: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Update chroots

import os
from glob import glob
from re import findall, DOTALL
from subprocess import check_output as call
from time import time


class DebomaticModule_UpdateChroots:

    def __init__(self):
        pass

    def __architecture(self, args):
        architecture = False
        if (args.opts.has_section('crossbuild') and
                args.opts.getboolean('crossbuild', 'crossbuild')):
            architecture = args.opts.get('crossbuild',
                                         'hostarchitecture')
        else:
            architecture = args.opts.get('debomatic', 'architecture')
            if architecture == 'system':
                b_arch = call(['dpkg-architecture', '-qDEB_BUILD_ARCH'])
                architecture = b_arch.strip().decode('utf-8')
        return architecture

    def __timestamp(self, filename, delta):
        current = time()
        try:
            with open(filename) as fd:
                last = float(fd.read().strip())
                if (current - last > delta):
                    with open(filename, 'w') as fd:
                        fd.write('{0}\n'.format(current))
                    return True
                else:
                    return False
        except FileNotFoundError:
            with open(filename, 'w') as fd:
                fd.write('{0}\n'.format(current))
            return False

    def periodic(self, args):
        if args.opts.has_section('updatechroots'):
            delta = args.opts.getint('updatechroots', 'days') * 24 * 60 * 60
            chroots = call(['/usr/bin/schroot', '--all-chroots', '-l'])
            arch = self.__architecture(args)
            chroots = findall(r'chroot:(\S+?-{0}-debomatic)\n?'.format(arch),
                              chroots.decode('utf-8'))
            for chroot in glob('/etc/schroot/chroot.d/*-debomatic-*'):
                with open(chroot) as fd:
                    content = fd.read()
                (name, path) = findall(r'\[(\S+?)\].*directory=(\S+)',
                                       content, DOTALL)[0]
                if name in chroots:
                    tag = os.path.join(path, '.debomatic')
                    if self.__timestamp(tag, delta):
                        call(['/usr/bin/sbuild-update', '-udcar',
                              '--arch={0}'.format(arch), name])



# Deb-o-Matic - Date stamp module
#
# Copyright (C) 2008-2009 David Futcher
# Copyright (C) 2008-2022 Luca Falavigna
#
# Authors: David Futcher <bobbo@ubuntu.com>
#          Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Displays build start and finish times into a file in the build directory

from datetime import datetime
from time import gmtime, mktime, strftime, time


class DebomaticModule_DateStamp:

    def __init__(self):
        self.ts = ''
        self.begin = ''
        self.end = ''
        self.uploader = ''

    def pre_build(self, args):
        self.ts = ('%(directory)s/pool/%(package)s/%(package)s.datestamp' %
                   {'directory': args.directory, 'package': args.package})
        with open(self.ts, 'w') as fd:
            if args.uploader:
                self.uploader = args.uploader
                fd.write('Uploaded by %s\n' % self.uploader)
            self.begin = gmtime(time())
            now = datetime.now().strftime('%A, %d %B %Y %H:%M')
            fd.write('Build started at %s\n' % now)

    def post_build(self, args):
        with open(self.ts, 'a') as fd:
            self.end = gmtime(time())
            now = datetime.now().strftime('%A, %d %B %Y %H:%M')
            elapsed = strftime('%H:%M:%S',
                               gmtime(mktime(self.end) - mktime(self.begin)))
            fd.write('Build finished at %s\n' % now)
            fd.write('Elapsed time: %s' % elapsed)



# Deb-o-Matic
#
# Copyright (C) 2007-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from fcntl import flock, LOCK_EX, LOCK_NB, LOCK_UN
from re import findall, DOTALL
from subprocess import Popen, PIPE

from Debomatic import dom
from .exceptions import DebomaticError


class GPG:

    def __init__(self, file):
        self._file = file
        self._error = None
        if dom.opts.has_option('gpg', 'gpg'):
            self._gpg = dom.opts.getboolean('gpg', 'gpg')
        else:
            self._gpg = False
        self._sig = None

    def __enter__(self):
        self._fd = open(self._file, 'r')
        flock(self._fd, LOCK_EX | LOCK_NB)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        flock(self._fd, LOCK_UN)
        self._fd.close()

    def _check_signature(self):
        if dom.opts.has_option('gpg', 'keyring'):
            self._keyring = dom.opts.get('gpg', 'keyring')
            if not os.path.isfile(self._keyring):
                self._keyring = None
                self._error = _('Keyring not found')
                raise DebomaticError
        gpgresult = Popen(['gpgv', '--keyring', self._keyring, self._file],
                          stderr=PIPE).communicate()[1]
        signature = findall(b'Good signature from "(.*) <(.*)>.*"', gpgresult)
        if signature:
            self._sig = signature[0]
        else:
            self._error = _('No valid signatures found')
            raise DebomaticError

    def _strip_signature(self):
        with open(self._file, 'r') as fd:
            data = fd.read()
        with open(self._file, 'w') as fd:
            try:
                fd.write(findall('\n\n(.*?)\n\n?-', data, DOTALL)[0])
            except IndexError:
                pass

    def check(self):
        if self._gpg:
            self._check_signature()
            if self._sig:
                self._strip_signature()
        return self._sig

    def error(self):
        return self._error



# Deb-o-Matic
#
# Copyright (C) 2007-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from glob import glob
from logging import debug, error, info
from re import findall
from signal import SIGTERM

from Debomatic import dom
from .build import Build
from .exceptions import DebomaticError
from .gpg import GPG


class Command():

    def __init__(self, commandfile):
        self.incoming = dom.opts.get('debomatic', 'incoming')
        self.cmdfile = os.path.join(self.incoming, commandfile)
        self._process_command()

    def _process_command(self):
        info(_('Processing %s') % os.path.basename(self.cmdfile))
        try:
            with GPG(self.cmdfile) as gpg:
                try:
                    self.uploader = gpg.check()
                except DebomaticError:
                    os.remove(self.cmdfile)
                    error(gpg.error())
                    return
        except IOError:
            return
        with open(self.cmdfile, 'r') as fd:
            cmd = fd.read()
        os.remove(self.cmdfile)
        cmd_nmu = findall(r'\s?binnmu\s+(\S+_\S+) (\S+) (\d+) "(.*)" (.*)',
                          cmd)
        cmd_builddep = findall(r'\s?builddep\s+(\S+_\S+) (\S+) (.*)', cmd)
        cmd_kill = findall(r'\s?kill\s+(\S+_\S+) (\S+)', cmd)
        cmd_porter = findall(r'\s?porter\s+(\S+_\S+) (\S+) (.*)', cmd)
        cmd_rebuild = findall(r'\s?rebuild\s+(\S+_\S+) (\S+) ?(\S*)', cmd)
        cmd_rm = findall(r'\s?rm\s+(.*)', cmd)
        if cmd_nmu:
            self._process_binnmu(cmd_nmu)
        if cmd_builddep:
            self._process_builddep(cmd_builddep)
        if cmd_kill:
            self._process_kill(cmd_kill)
        if cmd_porter:
            self._process_porter(cmd_porter)
        if cmd_rebuild:
            self._process_rebuild(cmd_rebuild)
        if cmd_rm:
            self._process_rm(cmd_rm)

    def _process_binnmu(self, packages):
        debug(_('Performing a binNMU build'))
        for _package in packages:
            package = _package[0].split('_')
            distribution = _package[1]
            binnmu = _package[2]
            changelog = _package[3]
            maintainer = _package[4]
            b = Build(package=package, distribution=distribution,
                      binnmu=(binnmu, changelog), maintainer=maintainer,
                      uploader=self.uploader)
            if dom.pool.schedule(b.run):
                debug(_('Thread for %s scheduled') % '_'.join(package))

    def _process_builddep(self, packages):
        debug(_('Performing a package rebuild with extra dependencies'))
        for _package in packages:
            package = _package[0].split('_')
            distribution = _package[1]
            extrabd = [x.strip() for x in _package[2].split(',')]
            b = Build(package=package, distribution=distribution,
                      extrabd=extrabd, uploader=self.uploader)
            if dom.pool.schedule(b.run):
                debug(_('Thread for %s scheduled') % '_'.join(package))

    def _process_kill(self, builds):
        debug(_('Killing build task'))
        for _build in builds:
            package, version = _build[0].split('_')
            distribution = _build[1]
            for task in dom.buildqueue:
                if task.match(package, version, distribution):
                    pid = task.get_pid()
                    if pid:
                        os.kill(pid, SIGTERM)
                        debug(_('Build killed for %(package)s_'
                                '%(version)s in %(dist)s') %
                              {'package': package, 'version': version,
                               'distribution': distribution})

    def _process_porter(self, packages):
        debug(_('Performing a porter build'))
        for _package in packages:
            package = _package[0].split('_')
            distribution = _package[1]
            maintainer = _package[2]
            b = Build(package=package, distribution=distribution,
                      maintainer=maintainer, uploader=self.uploader)
            if dom.pool.schedule(b.run):
                debug(_('Thread for %s scheduled') % '_'.join(package))

    def _process_rebuild(self, packages):
        debug(_('Performing a package rebuild'))
        for _package in packages:
            package = _package[0].split('_')
            distribution = _package[1]
            origin = _package[2] if _package[2] else distribution
            b = Build(package=package, distribution=distribution,
                      origin=origin, uploader=self.uploader)
            if dom.pool.schedule(b.run):
                debug(_('Thread for %s scheduled') % '_'.join(package))

    def _process_rm(self, filesets):
        debug(_('Removing files'))
        for files in filesets:
            for pattern in files.split():
                pattern = os.path.basename(pattern)
                for absfile in glob(os.path.join(self.incoming, pattern)):
                    debug(_('Removing %s') % absfile)
                    os.remove(absfile)



# Deb-o-Matic
#
# Copyright (C) 2011-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from concurrent.futures import as_completed, ThreadPoolExecutor
from atexit import register as on_exit
from fcntl import flock, LOCK_EX, LOCK_NB, LOCK_UN
from hashlib import sha256
from logging import basicConfig as log, error, getLogger, info
from signal import signal, SIGINT, SIGTERM
from sys import stdin, stdout, stderr
from threading import Timer as _Timer
from time import sleep

from Debomatic import dom
from .exceptions import DebomaticError


class Process:

    def __init__(self):
        pass

    def _daemonize(self):
        try:
            pid = os.fork()
            if pid > 0:
                exit()
        except OSError as e:
            error(_('Error entering daemon mode: %s') % e.strerror)
            exit()
        os.chdir('/')
        os.setsid()
        os.umask(0)
        stdout.flush()
        stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'a+')
        se = open(os.devnull, 'a+')
        os.dup2(si.fileno(), stdin.fileno())
        os.dup2(so.fileno(), stdout.fileno())
        os.dup2(se.fileno(), stderr.fileno())
        on_exit(self._quit)
        old_log = getLogger()
        if old_log.handlers:
            for handler in old_log.handlers:
                old_log.removeHandler(handler)
        log(filename=self.logfile, level=self.loglevel,
            format='%(asctime)s %(levelname)-8s %(message)s')
        self._set_pid()

    def _get_pid(self):
        self.pidfile = ('/var/run/debomatic-%s' %
                        self._sha256(self.incoming))
        try:
            with open(self.pidfile, 'r') as fd:
                self.pid = int(fd.read().strip())
        except (IOError, ValueError):
            self.pid = None

    def _lock(self, wait=False):
        self.fd = None
        self.lockfile = ('/var/run/debomatic-%s.lock' %
                         self._sha256(self.incoming))
        try:
            self.fd = open(self.lockfile, 'w')
            flags = LOCK_EX if wait else LOCK_EX | LOCK_NB
            flock(self.fd, flags)
        except (OSError, IOError) as ex:
            if self.fd:
                self.fd.close()
            raise ex

    def _notify_systemd(self):
        try:
            import systemd.daemon
            systemd.daemon.notify('READY=1')
        except (ImportError, SystemError):
            pass

    def _quit(self, signum=None, frame=None):
        info(_('Waiting for threads to complete...'))
        dom.periodic_event.cancel()
        dom.pool.shutdown()
        self.mod_sys.execute_hook('on_quit')
        self._unlock()
        os.unlink(self.pidfile)
        exit()

    def _set_pid(self):
        self.pidfile = ('/var/run/debomatic-%s' %
                        self._sha256(self.incoming))
        pid = str(os.getpid())
        with open(self.pidfile, 'w+') as fd:
            fd.write('%s\n' % pid)

    def _sha256(self, value):
        lock_sha = sha256()
        lock_sha.update(value.encode('utf-8'))
        return lock_sha.hexdigest()

    def _unlock(self):
        if self.fd:
            flock(self.fd, LOCK_UN)
            self.fd.close()
            self.fd = None
        if os.path.isfile(self.lockfile):
            os.unlink(self.lockfile)

    def shutdown(self):
        self._get_pid()
        if not self.pid:
            return
        info(_('Waiting for threads to complete...'))
        try:
            dom.periodic_event.cancel()
            os.kill(self.pid, SIGTERM)
            self._lock(wait=True)
        except OSError as err:
            err = str(err)
            if err.find('No such process') > 0:
                if os.path.exists(self.pidfile):
                    os.unlink(self.pidfile)
            else:
                error(err)

    def startup(self):
        try:
            self._lock()
        except (OSError, IOError):
            error(_('Another instance is running, aborting'))
            raise DebomaticError
        self._set_pid()
        signal(SIGINT, self._quit)
        signal(SIGTERM, self._quit)
        if self.daemonize:
            self._daemonize()
        self._notify_systemd()
        dom.periodic_event.start()
        self.launcher()


class ModulePool:

    def __init__(self, workers=1):
        self._jobs = {}
        self._pool = ThreadPoolExecutor(workers)

    def _launch(self, func, hook, dependencies):
        if dependencies:
            for dependency in dependencies:
                while True:
                    if dependency in self._jobs.keys():
                        self._jobs[dependency].result()
                        break
                    else:
                        sleep(0.1)
        func(hook)

    def schedule(self, func, hook):
        innerfunc, args, module, hookname, dependencies = hook
        job = self._pool.submit(self._launch, func, hook, dependencies)
        self._jobs[module] = job

    def shutdown(self):
        for job in as_completed([self._jobs[j] for j in self._jobs]):
            job.result()
        self._pool.shutdown()


class ThreadPool:

    def __init__(self, workers=1):
        self._jobs = []
        self._pool = ThreadPoolExecutor(workers)

    def _finish(self, job):
        try:
            self._jobs.remove(job)
        except ValueError:
            pass
        try:
            e = job.exception()
            if e:
                raise e
        except Exception as e:
            error(str(e), exc_info=True)

    def schedule(self, func):
        job = self._pool.submit(func)
        job.add_done_callback(self._finish)
        self._jobs.append(job)

    def shutdown(self):
        for job in as_completed(self._jobs):
            job.result()
        self._pool.shutdown()


class Timer(_Timer):

    def run(self):
        while not self.finished.is_set():
            self.finished.wait(self.interval)
            if not self.finished.is_set():
                self.function(*self.args, **self.kwargs)
        self.finished.set()



# Deb-o-Matic - RemovePackages module
#
# Copyright (C) 2018-2022 Luca Falavigna
#
# Authors: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Remove packages built long ago

import os
from shutil import rmtree
from time import time


class DebomaticModule_RemovePackages:

    def __init__(self):
        pass

    def periodic(self, args):
        ctime = time()
        if args.opts.has_section('removepackages'):
            delta = args.opts.getint('removepackages', 'days') * 24 * 60 * 60
            if os.path.isdir(args.directory):
                for suite in os.listdir(args.directory):
                    element = os.path.join(args.directory, suite)
                    pool = os.path.join(element, 'pool')
                    if os.path.isdir(pool):
                        for package in os.listdir(pool):
                            package = os.path.join(pool, package)
                            if os.path.isdir(package):
                                ptime = os.stat(package).st_mtime
                                if ptime + delta < ctime:
                                    rmtree(package)
                        if not os.listdir(pool):
                            rmtree(pool)
                    if os.path.isdir(element):
                        if not [f for f in os.listdir(element)
                                if f != 'dists' and f != 'logs']:
                            rmtree(element)



# Deb-o-Matic - Contents module
#
# Copyright (C) 2009 Alessio Treglia
# Copyright (C) 2010-2022 Luca Falavigna
#
# Authors: Alessio Treglia <quadrispro@ubuntu.com>
#          Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Store debc output in the pool directory.

import os
from subprocess import call


class DebomaticModule_Contents:

    def __init__(self):
        self.debc = '/usr/bin/debc'

    def post_build(self, args):
        if not args.success:
            return
        changes_file = None
        resultdir = os.path.join(args.directory, 'pool', args.package)
        contents_file = os.path.join(resultdir, args.package) + '.contents'
        if args.hostarchitecture:
            architecture = args.hostarchitecture
        else:
            architecture = args.architecture
        for filename in os.listdir(resultdir):
            if filename.endswith('.changes'):
                changes_file = os.path.join(resultdir, filename)
                break
        if changes_file and os.access(self.debc, os.X_OK):
            with open(contents_file, 'w') as fd:
                call([self.debc, '-a%s' % architecture, changes_file],
                     stdout=fd)



# Deb-o-Matic - Mailer module
#
# Copyright (C) 2010 Alessio Treglia
# Copyright (C) 2015-2022 Luca Falavigna
#
# Authors: Alessio Treglia <quadrispro@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Send a reply to the uploader once the build has finished.

import os
from glob import glob
from re import findall, DOTALL
from smtplib import SMTP
from email.parser import Parser


class DebomaticModule_Mailer:

    def __init__(self):
        self.after = ['Lintian']

    def write_reply(self, template, buildlog, package):
        lintianoutput = 'No log'
        with open(template, 'r') as fd:
            if self.lintian:
                lintfile = os.path.join(self.resultdir, '%s.lintian' % package)
                if os.path.isfile(lintfile):
                    with open(lintfile, 'r') as lintfd:
                        lintianoutput = lintfd.read()
            params = {'sender': self.sender,
                      'uploader': self.uploader,
                      'package': package,
                      'buildlog': buildlog,
                      'lintian': lintianoutput}
            reply = fd.read() % params
        msg = Parser().parsestr(reply)
        return msg.as_string().encode('utf-8')

    def post_build(self, args):
        if not args.uploader:
            return
        template = None
        self.uploader = args.uploader
        self.resultdir = os.path.join(args.directory, 'pool', args.package)
        if args.opts.has_section('mailer'):
            self.sender = args.opts.get('mailer', 'sender')
            self.server = args.opts.get('mailer', 'server')
            self.port = args.opts.getint('mailer', 'port')
            self.tls = args.opts.getboolean('mailer', 'tls')
            self.auth = args.opts.getboolean('mailer', 'authrequired')
            self.user = args.opts.get('mailer', 'user')
            self.passwd = args.opts.get('mailer', 'passwd')
            self.success = args.opts.get('mailer', 'success')
            self.failure = args.opts.get('mailer', 'failure')
            self.lintian = args.opts.getboolean('mailer', 'lintian')
        for filename in os.listdir(self.resultdir):
            if filename.endswith('.changes'):
                template = self.success
                break
        if not template:
            template = self.failure
        try:
            if args.hostarchitecture:
                architecture = args.hostarchitecture
            else:
                architecture = args.architecture
            bp = glob(os.path.join(args.directory, 'pool', args.package,
                                   '*_%s.build' % architecture))[0]
            with open(bp, 'r', encoding='utf8') as fd:
                if args.success:
                    data = findall('dpkg-buildpackage\n(.*)?\nBuild finished ',
                                   fd.read(), DOTALL)
                else:
                    data = findall('(.*)?\n┌─*┐\n│ Cleanup', fd.read(), DOTALL)
            try:
                if len(data[0]) > 20:
                    log = '\n'.join(data[0].split('\n')[-21:-1])
                else:
                    log = '\n'.join(data[0].split('\n'))
            except IndexError:
                return
            msg = self.write_reply(template, log, args.package)
            self.smtp = SMTP(self.server, self.port)
            self.smtp.ehlo()
            if args.opts.has_option('mailer', 'tls'):
                if args.opts.getboolean('mailer', 'tls'):
                    self.smtp.starttls()
            if self.auth:
                self.smtp.login(self.user, self.passwd)
            self.smtp.sendmail(self.sender, self.uploader, msg)
            self.smtp.quit()
        except Exception as e:
            raise e



# Deb-o-Matic - Package build path cleaner
#
# Copyright (C) 2009 Alessio Treglia
# Copyright (C) 2009-2022 Luca Falavigna
#
# Authors: Alessio Treglia <quadrispro@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Delete all files generated by previous build.

import os


class DebomaticModule_BuildCleaner:

    def __init__(self):
        self.after = ['Lintian', 'Piuparts']
        self.before = ['Repository']

    def pre_build(self, args):
        if args.opts.has_section('buildcleaner'):
            if args.opts.getboolean('buildcleaner', 'testbuild'):
                return
        exts_to_clean = ['.deb', '.ddeb', '.gz', '.bz2', '.xz', '.dsc',
                         '.changes', '.build', '.buildlog', '.buildinfo',
                         '.contents', '.lintian', '.piuparts', '.blhc',
                         '.autopkgtest', '.udeb']
        pkg_build_path = ('%(directory)s/pool/%(package)s' %
                          {'directory': args.directory,
                           'package': args.package})
        for filename in os.listdir(pkg_build_path):
            name, ext = os.path.splitext(filename)
            if ext in exts_to_clean:
                os.remove(os.path.join(pkg_build_path, filename))

    def post_build(self, args):
        exts_to_clean = ['.deb', '.ddeb', '.udeb', '.changes',
                         '.gz', '.bz2', '.xz', '.dsc']
        pkg_build_path = ('%(directory)s/pool/%(package)s' %
                          {'directory': args.directory,
                           'package': args.package})
        if args.opts.has_section('buildcleaner'):
            if args.opts.getboolean('buildcleaner', 'testbuild'):
                for filename in os.listdir(pkg_build_path):
                    name, ext = os.path.splitext(filename)
                    if ext in exts_to_clean:
                        os.remove(os.path.join(pkg_build_path, filename))



# Deb-o-Matic
#
# Copyright (C) 2007-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from argparse import ArgumentParser
from configparser import ConfigParser
from fcntl import flock, LOCK_EX, LOCK_NB, LOCK_UN
from logging import basicConfig as log, debug, error, getLogger, warning
from logging import ERROR, WARNING, INFO, DEBUG
from time import sleep

from Debomatic import dom
from .build import Build
from .commands import Command
from .configuration import Parser
from .exceptions import DebomaticConffileError, DebomaticError
from .modules import Module
from .process import Process, ThreadPool, Timer


class Debomatic(Parser, Process):

    def __init__(self):
        self.daemonize = True
        self.oneshot = False
        self.setlog('%(levelname)s: %(message)s')
        self.conffile = None
        dom.opts = ConfigParser()
        dom.dists = ConfigParser()
        parser = ArgumentParser()
        parser.add_argument('-c', '--configfile', metavar='file', type=str,
                            nargs=1, help='configuration file for Deb-o-Matic')
        parser.add_argument('-i', '--interactive', action='store_true',
                            help='launch Deb-o-Matic in interactive mode')
        parser.add_argument('-o', '--oneshot', metavar='file', type=str,
                            nargs=1, help='launch Deb-o-Matic for one build')
        parser.add_argument('-q', '--quit', action='store_true',
                            help='terminate Deb-o-Matic processes')
        args = parser.parse_args()
        if os.getuid():
            error(_('You must run Deb-o-Matic as root'))
            exit(1)
        if args.configfile:
            self.conffile = args.configfile[0]
        if args.interactive:
            self.daemonize = False
        if args.oneshot:
            self.oneshot = args.oneshot[0]
        try:
            self.parse_configfiles()
        except DebomaticConffileError:
            exit(1)
        self.incoming = dom.opts.get('debomatic', 'incoming')
        if not os.path.isdir(self.incoming):
            error(_('Unable to access %s directory') % self.incoming)
            exit(1)
        dom.pool = ThreadPool(dom.opts.getint('debomatic', 'threads'))
        dom.periodic_event = Timer(dom.opts.getint('debomatic', 'interval'),
                                   self.periodic_event)
        dom.buildqueue = []
        dom.chroots = {}
        self.logfile = dom.opts.get('debomatic', 'logfile')
        if args.quit:
            self.shutdown()
            exit()
        self.setlog('%(levelname)s: %(message)s',
                    dom.opts.get('debomatic', 'loglevel'))
        self.mod_sys = Module()
        self.mod_sys.execute_hook('on_start')
        try:
            self.startup()
        except DebomaticError:
            error(_('Another instance is running, aborting'))
            exit(1)

    def launcher(self):
        if self.oneshot:
            self.queue_files([self.oneshot])
        else:
            self.queue_files()
            if dom.opts.getboolean('debomatic', 'inotify'):
                try:
                    self.launcher_inotify()
                except ImportError:
                    self.launcher_timer()
            else:
                self.launcher_timer()

    def launcher_inotify(self):
        import pyinotify

        class PE(pyinotify.ProcessEvent, Debomatic):

            def __init__(self, parent):
                self.parent = parent

            def process_IN_CLOSE_WRITE(self, event):
                if (event.name.endswith('.changes') or
                        event.name.endswith('commands')):
                    self.parent.queue_files([event.name])

        wm = pyinotify.WatchManager()
        notifier = pyinotify.Notifier(wm, PE(self))
        wm.add_watch(self.incoming, pyinotify.IN_CLOSE_WRITE)
        debug(_('Inotify loop started'))
        notifier.loop()

    def launcher_timer(self):
        debug(_('Timer loop started'))
        while True:
            sleep(dom.opts.getint('debomatic', 'sleep'))
            self.queue_files()

    def queue_files(self, filelist=None):
        if filelist:
            filelist = [os.path.join(self.incoming, os.path.basename(f))
                        for f in filelist]
            for f in filelist:
                if not os.path.isfile(f):
                    filelist.remove(f)
        else:
            try:
                filelist = os.listdir(self.incoming)
            except OSError:
                error(_('Unable to access %s directory') % self.incoming)
                exit(1)
        for filename in filelist:
            if filename.endswith('.changes'):
                with open(os.path.join(self.incoming, filename)) as fd:
                    try:
                        flock(fd, LOCK_EX | LOCK_NB)
                        flock(fd, LOCK_UN)
                    except IOError:
                        pass
                    else:
                        b = Build(changesfile=filename)
                        dom.pool.schedule(b.run)
                        debug(_('Thread for %s scheduled') % filename)
            elif filename.endswith('.commands'):
                Command(filename)
        if self.oneshot:
            self.shutdown()

    def setlog(self, fmt, level='info'):
        loglevels = {'error': ERROR,
                     'warning': WARNING,
                     'info': INFO,
                     'debug': DEBUG}
        level = level.lower()
        if level not in loglevels:
            warning(_('Log level not valid, defaulting to "info"'))
            level = 'info'
        self.loglevel = loglevels[level]
        old_log = getLogger()
        if old_log.handlers:
            for handler in old_log.handlers:
                old_log.removeHandler(handler)
        log(level=self.loglevel, format=fmt)

    def periodic_event(self):
        self.mod_sys.args.opts = dom.opts
        self.mod_sys.args.directory = self.incoming
        self.mod_sys.execute_hook('periodic')



# Deb-o-Matic
#
# Copyright (C) 2008-2009 David Futcher
# Copyright (C) 2012-2022 Luca Falavigna
#
# Author: David Futcher <bobbo@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from glob import glob
from logging import debug
from re import findall
from sys import path
from toposort import toposort_flatten as toposort

from Debomatic import dom
from .process import ModulePool


class ModuleArgs():

    def __init__(self):
        self.opts = dom.opts
        self.action = None
        self.architecture = None
        self.directory = None
        self.distribution = None
        self.dists = None
        self.dsc = None
        self.files = None
        self.package = None
        self.success = False
        self.uploader = None
        self.hostarchitecture = None


class Module():

    def __init__(self):
        self.args = ModuleArgs()
        self._use_modules = False
        if dom.opts.has_option('modules', 'modules'):
            if dom.opts.getboolean('modules', 'modules'):
                self._use_modules = True
        if dom.opts.has_option('modules', 'path'):
            mod_path = dom.opts.get('modules', 'path')
            for p in mod_path.split(':'):
                path.append(p)
        else:
            self._use_modules = False
        modules = set()
        try:
            for p in mod_path.split(':'):
                modules |= set([os.path.splitext(os.path.basename(m))[0]
                               for m in glob(os.path.join(p, '*.py'))])
        except OSError:
            self._use_modules = False
        if not modules:
            self._use_modules = False
        if self._use_modules:
            self._instances = {}
            for module in modules:
                try:
                    _class = 'DebomaticModule_%s' % module
                    _mod = __import__(module)
                    self._instances[module] = getattr(_mod, _class)()
                    self._instances[module]._modulename = module
                    self._instances[module]._disabled = False
                    self._instances[module]._depends = set()
                    self._instances[module]._after = set()
                    debug(_('Module %s loaded') % module)
                except (NameError, SyntaxError):
                    pass
            self._set_relationships()
            self._set_blacklisted()
            self._disable_modules()
            debug(_('Modules will be executed in this order: %s') %
                  ', '.join([m for m in self._sort_modules()
                             if not self._instances[m]._disabled]))

    def _disable_modules(self):
        for module in self._sort_modules():
            missing = set()
            missing_after = set()
            for dep in self._instances[module]._depends:
                if dep in self._instances:
                    if self._instances[dep]._disabled:
                        missing.add(dep)
                else:
                    missing.add(dep)
            for dep in self._instances[module]._after:
                if dep in self._instances:
                    if self._instances[dep]._disabled:
                        missing_after.add(dep)
                else:
                    missing_after.add(dep)
            if missing:
                self._instances[module]._disabled = True
                debug(_('%(mod)s module disabled, needs %(missing)s') %
                      {'mod': module, 'missing': ', '.join(missing)})
            if missing_after:
                for dep in missing_after:
                    self._instances[module]._after.remove(dep)

    def _launcher(self, hook):
        func, args, module, hookname, dependencies = hook
        debug(_('Executing hook %(hook)s from module %(mod)s') %
              {'hook': hookname, 'mod': module})
        func(args)

    def _set_blacklisted(self):
        blist = set()
        if dom.opts.has_option('modules', 'blacklist'):
            _blacklist = dom.opts.get('modules', 'blacklist')
            blist = set(_blacklist.split())
        for module in self._instances:
            if module in blist:
                self._instances[module]._disabled = True
                debug(_('Module %s is blacklisted') % module)

    def _set_relationships(self):
        for module in self._instances:
            try:
                deps = getattr(self._instances[module], 'dependencies')
                if deps:
                    for dep in deps:
                        self._instances[module]._depends.add(dep)
            except AttributeError:
                pass
            try:
                afters = getattr(self._instances[module], 'after')
                if afters:
                    for after in afters:
                        self._instances[module]._after.add(after)
            except AttributeError:
                pass
            try:
                befores = getattr(self._instances[module], 'before')
                if befores:
                    for before in befores:
                        if before in self._instances:
                            self._instances[before]._after.add(module)
            except AttributeError:
                pass
            try:
                if getattr(self._instances[module], 'first'):
                    deps = (self._instances[module]._depends.union(
                            self._instances[module]._after))
                    if deps:
                        self._instances[module]._disabled = True
                        debug(_('Cannot execute %(mod)s as %(order)s module, '
                                'dependencies found: %(deps)s')
                              % {'mod': module, 'order': 'first',
                                 'deps': ', '.join(deps)})
                    else:
                        for instance in self._instances:
                            self._instances[instance]._after.add(module)
            except AttributeError:
                pass
            try:
                if getattr(self._instances[module], 'last'):
                    deps = [m for m in self._instances
                            if module in self._instances[m]._after or
                            module in self._instances[m]._depends]
                    if deps:
                        self._instances[module]._disabled = True
                        debug(_('Cannot execute %(mod)s as %(order)s module, '
                                'dependencies found: %(deps)s')
                              % {'mod': module, 'order': 'last',
                                 'deps': ', '.join(deps)})
                    else:
                        for instance in self._instances:
                            self._instances[module]._after.add(instance)
            except AttributeError:
                pass
            if module in self._instances[module]._depends:
                self._instances[module]._depend.remove(module)
            if module in self._instances[module]._after:
                self._instances[module]._after.remove(module)

    def _sort_modules(self):
        modules = {}
        for instance in self._instances:
            if not self._instances[instance]._disabled:
                _deps = self._instances[instance]._depends
                _afters = self._instances[instance]._after
                modules[instance] = _deps.union(_afters)
        try:
            return [m for m in toposort(modules) if m in self._instances]
        except ValueError as e:
            circular = findall(r'.*?\(\'?(\S+?)\'?,', e.args[0])
            for instance in circular:
                self._instances[instance]._disabled = True
            debug(_('Circular dependencies found, disabled modules: %s')
                  % ', '.join(circular))
            return self._sort_modules()

    def execute_hook(self, hook):
        hooks = []
        if self._use_modules:
            for module in self._sort_modules():
                if self._instances[module]._disabled:
                    continue
                dependencies = set()
                instance = self._instances[module]
                for dep in instance._depends.union(instance._after):
                    if dep in self._instances:
                        if (dep in instance._after and
                                self._instances[dep]._disabled):
                            continue
                        try:
                            if getattr(self._instances[dep], hook):
                                dependencies.add(dep)
                        except AttributeError:
                            continue
                try:
                    hooks.append((getattr(instance, hook),
                                 self.args, module, hook, dependencies))
                except AttributeError:
                    pass
            if dom.opts.has_option('modules', 'threads'):
                workers = dom.opts.getint('modules', 'threads')
            else:
                workers = 1
            debug(_('%s hooks launched') % hook)
            modulepool = ModulePool(workers)
            for hk in hooks:
                modulepool.schedule(self._launcher, hk)
            modulepool.shutdown()
            debug(_('%s hooks finished') % hook)



# Deb-o-Matic - SourceUpload module
#
# Copyright (C) 2014-2022 Luca Falavigna
#
# Authors: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Allows uploading source-only packages to Debian archive

import os
from re import match, sub


class DebomaticModule_SourceUpload:

    def post_build(self, args):
        if not args.success:
            return
        lines = []
        changesfile = None
        resultdir = os.path.join(args.directory, 'pool', args.package)
        for filename in os.listdir(resultdir):
            if filename.endswith('.changes'):
                changesfile = os.path.join(resultdir, filename)
                break
        if changesfile:
            with open(changesfile, 'r') as fd:
                cf = fd.read()
            for line in cf.split('\n'):
                if match(r'.*?\S+_\S+_\S+\.u?deb', line):
                    continue
                elif line.startswith('Architecture: '):
                    lines.append('Architecture: source')
                else:
                    lines.append(line)
            sourcecf = sub('_[^_]+?.changes',
                           '_sourceupload.changes', changesfile)
            with open(sourcecf, 'w') as sourcecf:
                sourcecf.write('\n'.join(lines))



# Deb-o-Matic
#
# Copyright (C) 2015-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.


class DebomaticConffileError(Exception):
    pass


class DebomaticError(Exception):
    pass



# Deb-o-Matic
#
# Copyright (C) 2007-2022 Luca Falavigna
# Copyright (C) 2010 Alessio Treglia
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from ast import literal_eval
from contextlib import contextmanager
from logging import debug, error, info
from pwd import getpwnam
from re import findall, match, search, sub
from shutil import chown, copy, copymode, move, rmtree
from subprocess import Popen, check_output
from tempfile import NamedTemporaryFile
from threading import Semaphore
from time import strftime
from urllib.request import Request, urlopen
from urllib.error import HTTPError, URLError

from Debomatic import dom
from .gpg import GPG
from .modules import Module
from .exceptions import DebomaticError


class BuildTask:

    def __init__(self, build, package, version, distribution, queue):
        self._build = build
        self._package = package
        self._version = version
        self._distribution = distribution
        self._queue = queue
        self._pid = 0

    def __enter__(self):
        for task in self._queue:
            if self.match(task._package, task._version, task._distribution):
                info(_('Build already scheduled for '
                       'package %(package)s_%(version)s in %(dist)s') %
                     {'package': self._package, 'version': self._version,
                      'dist': self._distribution})
                self._skip_removal()
                raise DebomaticError
        self._queue.append(self)
        return self

    def __exit__(self, type, value, tb):
        if self in self._queue:
            self._queue.remove(self)

    def _skip_removal(self):
        for pkgfile in set(self._build.files):
            if os.path.isfile(pkgfile):
                for build in [x._build for x in self._queue
                              if x._build != self]:
                    if pkgfile in build.files:
                        self._build.files.remove(pkgfile)
                        debug(_('Skipping removal of file %s') % pkgfile)

    def get_pid(self):
        return self._pid

    @contextmanager
    def set_pid(self, pid):
        self._pid = pid
        try:
            yield
        finally:
            self._pid = 0

    def match(self, package, version, distribution):
        return (self._package == package and
           self._version == version and
           self._distribution == distribution)


class Build:

    def __init__(self, changesfile=None, package=None, distribution=None,
                 binnmu=None, extrabd=None, maintainer=None, origin=None,
                 uploader=None):
        self.uid = None
        self.gid = None
        self.buildtask = None
        self.changesfile = changesfile
        self.package = package
        self.suite = None
        self.distribution = distribution
        self.binnmu = binnmu
        self.extrabd = extrabd
        self.maintainer = maintainer
        self.origin = origin
        self.uploader = uploader
        self.files = set()
        self.incoming = dom.opts.get('debomatic', 'incoming')
        self.hostarchitecture = None
        self.dpr = False

    def _build(self):
        try:
            self.uid = getpwnam(dom.opts.get('debomatic', 'builduser')).pw_uid
            self.gid = getpwnam('sbuild').pw_gid
        except KeyError:
            self.uid = 0
            self.gid = 0
        self._parse_distribution()
        if self.distribution in dom.opts.get('distributions', 'blacklist'):
            self._remove_files()
            error(_('Distribution %s is disabled') % self.distribution)
            raise DebomaticError
        if dom.opts.has_section('crossbuild'):
            if dom.opts.getboolean('crossbuild', 'crossbuild'):
                self.hostarchitecture = dom.opts.get('crossbuild',
                                                     'hostarchitecture')
        try:
            if self.changesfile:
                package = os.path.basename(self.changesfile).split('_')[0]
                version = os.path.basename(self.changesfile).split('_')[1]
            else:
                package, version = self.package
            with BuildTask(self, package, version, self.suite,
                           dom.buildqueue) as bt:
                self.buildtask = bt
                self._fetch_files()
                self._setup_chroot()
                self._build_package()
        except DebomaticError:
            self._remove_files()

    def _build_package(self):
        uploader_email = ''
        packageversion = os.path.splitext(os.path.basename(self.dscfile))[0]
        builddir = os.path.join(self.poolpath, 'pool', packageversion)
        self._makedirs(builddir, True)
        if self.uploader:
            uploader_email = self.uploader[1].decode('utf-8')
        architecture = dom.opts.get('debomatic', 'architecture')
        if architecture == 'system':
            b_arch = check_output(['dpkg-architecture', '-qDEB_BUILD_ARCH'])
            architecture = b_arch.strip().decode('utf-8')
        mod = Module()
        mod.args.architecture = architecture
        mod.args.directory = self.poolpath
        mod.args.distribution = self.distribution
        mod.args.dists = dom.dists
        mod.args.dsc = self.dscfile
        mod.args.files = self.files
        mod.args.package = packageversion
        mod.args.uploader = uploader_email
        mod.args.hostarchitecture = self.hostarchitecture
        mod.execute_hook('pre_build')
        info(_('Building %s') % os.path.basename(self.dscfile))
        command = ['sbuild', '-A', '-s', '-d', self.distribution,
                   '--arch=%s' % architecture, '--no-run-lintian', '-c',
                   '%s-%s-debomatic' % (self.distribution, architecture),
                   self.dscfile]
        if self.hostarchitecture:
            command.pop(5)
            command.insert(5, '--host=%s' % self.hostarchitecture)
            command.insert(6, '--add-depends=libc6-dev:%s' %
                           self.hostarchitecture)
        suite = dom.dists.get(self.distribution, 'suite')
        if self.distribution != suite:
            command.insert(-1, '--build-dep-resolver=aspcud')
            command.insert(-1, ('--aspcud-criteria=-removed,-changed,-new,'
                                '-count(solution,APT-Release:=/%s/)' %
                                self.distribution))
        if self.distribution != self.suite:
            command.insert(-1, '--build-dep-resolver=aspcud')
            command.insert(-1, ('--aspcud-criteria=-removed,-changed,-new,'
                                '-count(solution,APT-Release:=/%s/)' %
                                self.suite))
        if self.extrabd:
            for extrabd in self.extrabd:
                command.insert(-1, '--add-depends=%s' % extrabd)
            command.insert(-1, '--build-dep-resolver=aspcud')
        if self.changesfile:
            with open(self.upload, 'r') as fd:
                data = fd.read()
            for file in findall(r'\s\w{32}\s\d+\s\S+\s\S+\s(.*)', data):
                if '.orig.' in file:
                    command.insert(-1, '--force-orig-source')
                    break
            try:
                command.insert(-1, '--debbuildopt=-v%s~' %
                               findall(r' \S+ \((\S+)\) \S+; ', data)[-1])
            except IndexError:
                pass
            with open(os.path.join(self.incoming, self.changesfile)) as fd:
                data = fd.read()
            for resolver in findall(r'Debomatic-Resolver: (\S+)', data):
                command.insert(-1, '--build-dep-resolver=%s' % resolver)
        if self.binnmu:
            command.insert(-1, '--binNMU=%s' % self.binnmu[0])
            command.insert(-1, '--make-binNMU=%s' % self.binnmu[1])
            command.insert(-1, '--no-arch-all')
            buildlog = '%s+b%s_%s.build' % (packageversion, self.binnmu[0],
                                            architecture)
        else:
            buildlog = '%s_%s.build' % (packageversion, architecture)
        if self.hostarchitecture:
            buildlog = sub(r'(.*_)\S+(\.build)',
                           '\\1%s\\2' % self.hostarchitecture, buildlog)
        if self.maintainer:
            command.remove('-A')
            command.remove('-s')
            command.insert(-1, '--maintainer=%s' % self.maintainer)
        ext = {'.gz': 'gzip', '.bz2': 'bzip2', '.xz': 'xz'}
        for file in self.files:
            if os.path.isfile(file):
                if findall(r'(.*\.debian\..*)', file):
                    try:
                        command.insert(-1, '--debbuildopt=-Z%s' %
                                       ext[os.path.splitext(file)[1]])
                    except IndexError:
                        pass
        for sbuildcommand in self._commands(self.distribution, architecture,
                                            packageversion):
            command.insert(-1, sbuildcommand)
        if self.dpr:
            if dom.opts.get('dpr', 'repository'):
                command.insert(-1, '--extra-repository=%s' %
                               (dom.opts.get('dpr', 'repository') %
                                {'dist': os.path.basename(self.poolpath)}))
        with open(os.devnull, 'w') as fd:
            try:
                ppath = os.path.join(self.poolpath, 'pool', packageversion)
                buildlink = os.path.join(ppath, '%s.buildlog' % packageversion)
                if os.path.exists(buildlink):
                    os.unlink(buildlink)
                os.symlink(buildlog, buildlink)
                process = Popen(command, stdout=fd, stderr=fd, cwd=ppath,
                                preexec_fn=self._demote(),
                                start_new_session=True)
                with self.buildtask.set_pid(process.pid):
                    process.wait()
                if process.returncode:
                    info(_("Build of %s failed") %
                         os.path.basename(self.dscfile))
                else:
                    info(_("Build of %s successful") %
                         os.path.basename(self.dscfile))
                    mod.args.success = True
            except OSError:
                error(_('Invocation of sbuild failed'))
        mod.execute_hook('post_build')
        self._remove_files()
        debug(_('Build of %s complete') % os.path.basename(self.dscfile))

    def _commands(self, distribution, architecture, packageversion):
        commands = []
        types = ('pre-build-commands', 'chroot-setup-commands',
                 'build-deps-failed-commands', 'starting-build-commands',
                 'build-failed-commands', 'finished-build-commands',
                 'chroot-cleanup-commands', 'post-build-commands')
        commandsdir = dom.opts.get('chroots', 'commands')
        if os.path.isdir(commandsdir):
            for type in types:
                if os.path.isdir(os.path.join(commandsdir, type)):
                    for command in os.listdir(os.path.join(commandsdir, type)):
                        commandfile = os.path.join(commandsdir, type, command)
                        if os.access(commandfile, os.X_OK):
                            commands.append('--%s=%s %s %s %s' %
                                            (type, commandfile,
                                             packageversion, distribution,
                                             architecture))
        if commands:
            commands.append('--log-external-command-output')
            commands.append('--log-external-command-error')
        return commands

    def _demote(self):
        def result():
            os.setgid(self.gid)
            os.setuid(self.uid)
        return result

    def _fetch_files(self):

        def _download_files(mirror, component, package, file, filepath):
            request = Request('%s/pool/%s/%s/%s/%s' % (mirror, component,
                              findall(r'^lib\S|^\S', package)[0],
                              package, file))
            try:
                debug(_('Requesting URL %s') % request.get_full_url())
                data = urlopen(request).read()
                with open(filepath, 'wb') as fd:
                    fd.write(data)
            except (HTTPError, URLError):
                pass

        dscfile = None
        components = dom.dists.get(self.origin, 'components').split()
        if self.changesfile:
            package = os.path.basename(self.changesfile).split('_')[0]
            for filename in self.files:
                if filename.endswith('.dsc'):
                    dscfile = filename
                    self.dscfile = os.path.join(self.incoming, dscfile)
                    break
            if not dscfile:
                self._remove_files()
                error(_('Bad .changes file: %s') % self.changesfile)
                raise DebomaticError
        else:
            package = self.package[0]
            version = sub(r'^\d+\:', '', self.package[1])
            dscfile = '%s_%s.dsc' % (package, version)
            if not dom.dists.has_section(self.origin):
                error(_('Distribution %s not configured') % self.distribution)
                raise DebomaticError
            self.dscfile = os.path.join(self.incoming, dscfile)
            if not os.path.isfile(self.dscfile):
                debug(_('Downloading missing %s') % dscfile)
                for component in components:
                    _download_files(dom.dists.get(self.origin, 'mirror'),
                                    component, package,
                                    dscfile, self.dscfile)
                    if os.path.isfile(self.dscfile):
                        break
        if self.dscfile and os.path.isfile(self.dscfile):
            self.files.add(self.dscfile)
            debug(_('File %s added') % self.dscfile)
        else:
            error(_('Unable to fetch %s') % dscfile)
            raise DebomaticError
        with open(self.dscfile, 'r') as fd:
            data = fd.read()
        for entry in findall(r'\s\w{32}\s\d+\s(\S+)', data):
            if not os.path.isfile(os.path.join(self.incoming, entry)):
                debug(_('Downloading missing %s') % entry)
                for component in dom.dists.get(self.origin,
                                               'components').split():
                    _download_files(dom.dists.get(self.origin, 'mirror'),
                                    component, package, entry,
                                    os.path.join(self.incoming, entry))
                    if os.path.isfile(os.path.join(self.incoming, entry)):
                        break
            if os.path.isfile(os.path.join(self.incoming, entry)):
                if not (os.path.join(self.incoming, entry)) in self.files:
                    entry = os.path.join(self.incoming, entry)
                    self.files.add(entry)
                    debug(_('File %s added') % entry)
            else:
                error(_('Unable to fetch %s') % entry)
                raise DebomaticError

    def _lock_chroot(self, chrootname):
        if chrootname not in dom.chroots:
            dom.chroots[chrootname] = Semaphore()
        dom.chroots[chrootname].acquire()

    def _map_distribution(self):
        if dom.opts.has_section('dpr'):
            if dom.opts.getboolean('dpr', 'dpr'):
                self.suite = self.distribution
                if match(r'%s-\S+-\S+' %
                         dom.opts.get('dpr', 'prefix'), self.suite):
                    self.dpr = True
                    self.distribution = '-'.join(
                        self.distribution.split('-')[2:])
                    if self.origin == self.suite:
                        self.origin = self.distribution
        if dom.opts.has_option('distributions', 'mapper'):
            try:
                mapper = literal_eval(dom.opts.get('distributions', 'mapper'))
            except SyntaxError:
                pass
            else:
                if self.distribution in mapper:
                    debug(_('%(mapped)s mapped as %(mapper)s') %
                          {'mapped': self.distribution,
                           'mapper': mapper[self.distribution]})
                    self.distribution = mapper[self.distribution]
                if self.origin and self.origin in mapper:
                    debug(_('%(mapped)s mapped as %(mapper)s') %
                          {'mapped': self.origin,
                           'mapper': mapper[self.origin]})
                    self.origin = mapper[self.origin]

    def _makedirs(self, directory, change_owner=False):
        os.makedirs(directory, exist_ok=True)
        if change_owner:
            chown(directory, user=self.uid)
            for root, dirs, files in os.walk(directory):
                try:
                    for name in dirs:
                        chown(os.path.join(root, name), user=self.uid)
                    for name in files:
                        chown(os.path.join(root, name), user=self.uid)
                except FileNotFoundError:
                    continue

    def _parse_distribution(self):
        if not self.distribution:
            try:
                with open(self.upload, 'r') as fd:
                    data = fd.read()
            except IOError:
                error(_('Unable to open %s') % self.upload)
                raise DebomaticError
            try:
                dist = findall(r'Distribution:\s+(\S+)', data)[0]
                self.distribution = dist.lower()
            except IndexError:
                error(_('Bad .changes file: %s') % self.upload)
                raise DebomaticError
        self.poolpath = os.path.join(self.incoming, self.distribution)
        self._map_distribution()
        if not dom.dists.has_section(self.distribution):
            error(_('Distribution %s not configured') % self.distribution)
            raise DebomaticError
        if self.origin:
            if not dom.dists.has_section(self.origin):
                error(_('Distribution %s not configured') % self.origin)
                raise DebomaticError
        else:
            self.origin = self.distribution
        if not self.suite:
            self.suite = self.distribution

    def _remove_files(self):
        for pkgfile in self.files:
            if os.path.isfile(pkgfile):
                os.remove(pkgfile)
                debug(_('File %s removed') % pkgfile)

    def _setup_chroot(self):
        action = None
        self.buildpath = os.path.join(self.incoming, self.distribution)
        self._makedirs(self.buildpath)
        self._makedirs(self.poolpath)
        architecture = dom.opts.get('debomatic', 'architecture')
        if architecture == 'system':
            b_arch = check_output(['dpkg-architecture', '-qDEB_BUILD_ARCH'])
            architecture = b_arch.strip().decode('utf-8')
        debootstrap = dom.opts.get('debomatic', 'debootstrap')
        chrootname = '%s-%s-debomatic' % (self.distribution, architecture)
        self._lock_chroot(chrootname)
        with open(os.devnull, 'w') as fd:
            chroots = check_output(['schroot', '-l'], stderr=fd)
        if not search('chroot:%s' % chrootname, chroots.decode()):
            action = 'create'
        mod = Module()
        mod.args.architecture = architecture
        mod.args.action = action
        mod.args.directory = self.buildpath
        mod.args.distribution = self.distribution
        mod.args.dists = dom.dists
        mod.execute_hook('pre_chroot')
        for d in ('logs', 'pool'):
            self._makedirs(os.path.join(self.buildpath, d))
        if action:
            profile = dom.opts.get('chroots', 'profile')
            if not os.path.isdir(os.path.join('/etc/schroot', profile)):
                error(_('schroot profile %s not found') % profile)
                self._unlock_chroot(chrootname)
                raise DebomaticError
            logfile = ('%s/logs/%s.%s' %
                       (self.buildpath, self.distribution,
                        strftime('%Y%m%d_%H%M%S')))
            target = dom.dists.get(self.distribution, 'suite')
            if target == self.distribution:
                pattern = '%s-%s-debomatic' % (self.distribution, architecture)
            else:
                pattern = '%s-%s-%s-debomatic' % (target, architecture,
                                                  self.distribution)
            if os.path.isdir(os.path.join(self.buildpath, self.distribution)):
                rmtree(os.path.join(self.buildpath, self.distribution))
            with open(logfile, 'w') as fd:
                try:
                    debug(_('Creating chroot %(dist)s-%(arch)s-debomatic') %
                          {'dist': self.distribution, 'arch': architecture})
                    components = ','.join(dom.dists.get(self.distribution,
                                                        'components').split())
                    command = ['sbuild-createchroot',
                               '--arch=%s' % architecture,
                               '--chroot-suffix=-debomatic',
                               '--debootstrap=%s' % debootstrap,
                               '--components=%s' % components, target,
                               os.path.join(self.buildpath, self.distribution),
                               dom.dists.get(self.distribution, 'mirror')]
                    if target != self.distribution:
                        command[2] = ('--chroot-suffix=-%s-debomatic' %
                                      self.distribution)
                    if dom.dists.has_option(self.distribution,
                                            'extrapackages'):
                        packages = dom.dists.get(self.distribution,
                                                 'extrapackages').split()
                        command.insert(-3, '--include=%s' % ','.join(packages))
                        packages = '--include=%s' % ','.join(packages)
                    process = Popen(command, stdout=fd, stderr=fd)
                    with self.buildtask.set_pid(process.pid):
                        process.wait()
                    if process.returncode:
                        error(_('Failed creating %(dist)s-%(arch)s-debomatic')
                              % {'dist': self.distribution,
                                 'arch': architecture})
                        mod.execute_hook('post_chroot')
                        self._unlock_chroot(chrootname)
                        raise DebomaticError
                except OSError:
                    error(_('Unable to launch sbuild-createchroot'))
                    mod.execute_hook('post_chroot')
                    self._unlock_chroot(chrootname)
                    raise DebomaticError
            if dom.dists.has_option(self.distribution, 'extramirrors'):
                with open(os.path.join(self.buildpath, self.distribution,
                                       'etc/apt/sources.list'), 'a') as fd:
                    fd.write(dom.dists.get(self.distribution, 'extramirrors'))
            if dom.opts.has_option('repository', 'pubring'):
                if os.path.isfile(dom.opts.get('repository', 'pubring')):
                    copy(dom.opts.get('repository', 'pubring'),
                         os.path.join(self.buildpath, self.distribution,
                                      'etc/apt/trusted.gpg.d/debomatic.gpg'))
            chroots = '/etc/schroot/chroot.d'
            for file in os.listdir(chroots):
                if file.startswith(pattern):
                    with NamedTemporaryFile(mode='w+', delete=False) as tmp:
                        with open(os.path.join(chroots, file)) as fd:
                            for line in fd:
                                if line.startswith('['):
                                    tmp.write('[%s-%s-debomatic]\n' %
                                              (self.distribution,
                                               architecture))
                                elif line.startswith('description'):
                                    tmp.write(line.replace(target,
                                              self.distribution))
                                elif line.startswith('profile'):
                                    tmp.write('profile=%s\n' % profile)
                                else:
                                    tmp.write(line)
                    copymode(fd.name, tmp.name)
                    move(tmp.name, fd.name)
            mod.args.success = True
            mod.execute_hook('post_chroot')
        self._unlock_chroot(chrootname)

    def _unlock_chroot(self, chrootname):
        if chrootname in dom.chroots:
            dom.chroots[chrootname].release()

    def run(self):
        if self.changesfile:
            self.upload = os.path.join(self.incoming, self.changesfile)
            self.files.add(self.upload)
            debug(_('File %s added') % self.upload)
            info(_('Processing %s') % self.changesfile)
            try:
                with open(self.upload, 'r') as fd:
                    data = fd.read()
            except IOError:
                error(_('Unable to open %s') % self.upload)
                raise DebomaticError
            try:
                for entry in findall(r'\s\w{32}\s\d+\s\S+\s\S+\s(.*)', data):
                    entry = os.path.join(self.incoming, entry)
                    self.files.add(entry)
                    debug(_('File %s added') % entry)
            except IndexError:
                error(_('Bad .changes file: %s') % self.upload)
                raise DebomaticError
            try:
                with GPG(self.upload) as gpg:
                    try:
                        self.uploader = gpg.check()
                    except DebomaticError:
                        self._remove_files()
                        error(gpg.error())
                        raise DebomaticError
                    try:
                        self._build()
                    except DebomaticError:
                        self._remove_files()
                        error(_('Build of %s failed') % self.changesfile)
            except (IOError, DebomaticError):
                pass
        else:
            try:
                self._build()
            except DebomaticError:
                self._remove_files()
                error(_('Build of %s failed') % '_'.join(self.package))



# Deb-o-Matic - Repository module
#
# Copyright (C) 2011-2022 Luca Falavigna
#
# Authors: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Generate local repository of built packages

import os
from fcntl import flock, LOCK_EX, LOCK_NB, LOCK_SH, LOCK_UN
from shutil import rmtree
from stat import S_IRWXU, S_IRGRP, S_IXGRP, S_IROTH, S_IXOTH
from subprocess import Popen, PIPE
from tempfile import mkdtemp, mkstemp


class DebomaticModule_Repository:

    def __init__(self):
        self.af = '/usr/bin/apt-ftparchive'
        self.gpg = '/usr/bin/gpg'

    def pre_build(self, args):
        self.update_repository(args)

    def post_build(self, args):
        self.update_repository(args)

    def pre_chroot(self, args):
        if args.action:
            self.update_repository(args)

    class Lock:

        def __init__(self, distribution, architecture):
            self._file = ('/var/run/debomatic-%s-%s.apt.lock' %
                          (distribution, architecture))
            self._skip = False

        def __enter__(self):
            self._fd = open(self._file, 'w')
            try:
                flock(self._fd, LOCK_EX | LOCK_NB)
            except IOError:
                flock(self._fd, LOCK_SH)
                self._skip = True
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            flock(self._fd, LOCK_UN)
            self._fd.close()

        def skip(self):
            return self._skip

    def update_repository(self, args):
        if args.opts.has_section('repository'):
            gpgkey = args.opts.get('repository', 'gpgkey')
            pubring = args.opts.get('repository', 'pubring')
        else:
            return
        if not os.access(self.af, os.X_OK):
            return
        if not os.access(self.gpg, os.X_OK):
            return
        distribution = os.path.basename(args.directory)
        archive = args.directory
        pool = os.path.join(archive, 'pool')
        distslink = os.path.join(archive, 'dists', distribution)
        if not os.path.islink(distslink):
            if not os.path.isdir(os.path.join(archive, 'dists')):
                os.makedirs(os.path.join(archive, 'dists'))
            tmpdir = mkdtemp(prefix='.', dir=os.path.join(archive, 'dists'))
            os.symlink(tmpdir, distslink)
        dists = mkdtemp(prefix='.', dir=os.path.join(archive, 'dists'))
        for arch in [a for a in (args.architecture,
                                 args.hostarchitecture) if a is not None]:
            packages = os.path.join(dists, 'main', 'binary-%s' % arch)
            packages_file = os.path.join(packages, 'Packages')
            arch_release_file = os.path.join(packages, 'Release')
            release_file = os.path.join(dists, 'Release')
            release_gpg = os.path.join(dists, 'Release.gpg')
            inrelease_gpg = os.path.join(dists, 'InRelease')
            for dir in (pool, packages):
                if not os.path.isdir(dir):
                    os.makedirs(dir)
            with self.Lock(distribution, arch) as lock:
                if not lock.skip():
                    with open(packages_file, 'w') as fd:
                        Popen([self.af, '-a', arch, 'packages', 'pool'],
                              stdout=fd, stderr=PIPE, cwd=archive).wait()
                    with open(arch_release_file, 'w') as fd:
                        fd.write('Origin: Deb-O-Matic\n')
                        fd.write('Label: Deb-O-Matic\n')
                        fd.write('Archive: %s\n' % distribution)
                        fd.write('Component: main\n')
                        fd.write('Architecture: %s\n' % arch)
                    with open(release_file, 'w') as fd:
                        afstring = 'APT::FTPArchive::Release::'
                        Popen([self.af, '-qq',
                               '-o', '%sOrigin=Deb-o-Matic' % afstring,
                               '-o', '%sLabel=Deb-o-Matic' % afstring,
                               '-o', '%sSuite=%s' % (afstring, distribution),
                               '-o', '%sArchitectures=%s' % (afstring, arch),
                               '-o', '%sComponents=main' % afstring,
                               'release',
                               'dists/%s' % os.path.basename(dists)],
                              stdout=fd, stderr=PIPE, cwd=archive).wait()
                    with open(release_file, 'r+') as fd:
                        data = fd.read()
                        fd.seek(0)
                        fd.write(data.replace('MD5Sum',
                                 'NotAutomatic: yes\nMD5Sum'))
                with open(release_gpg, 'w') as fd:
                    Popen([self.gpg, '--no-default-keyring', '--keyring',
                           pubring, '-u', gpgkey, '--yes', '-a', '-o', fd.name,
                           '-b', release_file], cwd=archive).wait()
                with open(inrelease_gpg, 'w') as fd:
                    Popen([self.gpg, '--no-default-keyring', '--keyring',
                           pubring, '-u', gpgkey, '--yes', '-a', '-o', fd.name,
                           '--clearsign', release_file], cwd=archive).wait()
        olddists = os.readlink(distslink)
        (tmp, tmplink) = mkstemp(prefix='.',
                                 dir=os.path.dirname(olddists))
        os.close(tmp)
        os.unlink(tmplink)
        os.symlink(dists, tmplink)
        os.chmod(dists, S_IRWXU | S_IRGRP | S_IXGRP |
                 S_IROTH | S_IXOTH)
        os.rename(tmplink, distslink)
        try:
            rmtree(olddists)
        except FileNotFoundError:
            pass



# Deb-o-Matic - PiuParts module
#
# Copyright (C) 2012 Leo Iannacone
# Copyright (C) 2015-2022 Luca Falavigna
#
# Authors: Leo Iannacone <l3on@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Check through piuparts if packages can be installed/upgraded/removed.

import os
from subprocess import call


class DebomaticModule_Piuparts:

    def __init__(self):
        self.piuparts = '/usr/sbin/piuparts'

    def post_build(self, args):
        if not args.success:
            return
        if not os.access(self.piuparts, os.X_OK):
            return
        if args.hostarchitecture:
            return
        if args.opts.has_section('piuparts'):
            options = args.opts.get('piuparts', 'options').strip().split()
        else:
            options = []
        distribution = args.distribution
        mirror = args.dists.get(distribution, 'mirror')
        schroot = '%s-%s-debomatic' % (distribution, args.architecture)
        resultdir = os.path.join(args.directory, 'pool', args.package)
        for filename in os.listdir(resultdir):
            if filename.endswith('.changes'):
                log = os.path.join(resultdir, args.package) + '.piuparts'
                with open(log, 'a') as fd:
                    cmd = [self.piuparts,
                           '-I', '/build/.*',
                           '-d', '%s' % distribution,
                           '-D', '%s' % mirror.split('/')[-1],
                           '--schroot=chroot:%s' % schroot,
                           os.path.join(resultdir, filename)]
                    for option in options:
                        cmd.insert(-1, option)
                    call(cmd, stdout=fd)
                break



#!/usr/bin/python3
# Deb-o-Matic
#
# Copyright (C) 2007-2022 Luca Falavigna
#
# Author: Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from distutils.core import setup
from distutils.command.install_data import install_data
from distutils.dir_util import mkpath
from distutils.file_util import copy_file
from glob import glob
from re import sub
from subprocess import call


for po in glob(os.path.join('po', '*.po')):
    mo = os.path.join('locale', os.path.splitext(os.path.basename(po))[0],
                      'LC_MESSAGES/debomatic.mo')
    if not os.path.isdir(os.path.dirname(mo)):
        os.makedirs(os.path.dirname(mo))
    call(['msgfmt', '-o', mo, po])


class SbuildCommands():

    def __init__(self):
        self.directories = ('build-deps-failed-commands',
                            'build-failed-commands',
                            'chroot-cleanup-commands',
                            'chroot-setup-commands',
                            'finished-build-commands',
                            'post-build-commands',
                            'pre-build-commands',
                            'starting-build-commands')

    def __enter__(self):
        for directory in self.directories:
            try:
                os.mkdir(os.path.join('sbuildcommands', directory))
            except FileExistsError:
                pass
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        for directory in self.directories:
            try:
                os.rmdir(os.path.join('sbuildcommands', directory))
            except OSError:
                pass


class InstallData(install_data):

    def run(self):
        call(['make', '-C', 'docs', 'latexpdf'])
        self.data_files.extend([('share/doc/debomatic',
                                 ['docs/_build/latex/Deb-o-Matic.pdf'])])
        self.install_files('etc')
        self.install_files('usr')
        self.install_files('modules', 'share/debomatic')
        with SbuildCommands():
            self.install_files('sbuildcommands', 'share/debomatic', True)
        self.install_files('locale', 'share')
        install_data.run(self)
        mkpath(os.path.join(self.install_dir, 'sbin'))
        if os.path.isabs(self.install_dir):
            install_dir = self.install_dir
        else:
            install_dir = os.path.join(os.getcwd(), self.install_dir)
        dstlink = os.path.join(install_dir, 'sbin/debomatic')
        if os.path.islink(dstlink):
            os.remove(dstlink)
        copy_file(os.path.join(install_dir, 'share/debomatic/debomatic'),
                  dstlink, link='sym')

    def install_files(self, rootdir, prefix='', empty=False):
        filelist = []
        for root, subFolders, files in os.walk(rootdir):
            dirlist = []
            for file in files:
                dirlist.append(os.path.join(root, file))
            if dirlist or empty:
                filelist.append((os.path.join(prefix, root), dirlist))
        if not prefix:
            orig_prefix = self.install_dir
            orig_data_files = self.data_files
            self.install_dir = sub('/*usr/*$', '/', self.install_dir)
            self.data_files = filelist
            install_data.run(self)
            self.install_dir = orig_prefix
            self.data_files = orig_data_files
        else:
            self.data_files.extend(filelist)


setup(name='debomatic',
      version='0.26',
      author='Luca Falavigna',
      author_email='dktrkranz@debian.org',
      description='Automatic build machine for Debian source packages',
      url='http://debomatic.github.io',
      license='GNU GPL',
      packages=['Debomatic'],
      data_files=[('share/debomatic', ['debomatic']),
                  ('share/man/man1', ['docs/debomatic.1'])],
      cmdclass={'install_data': InstallData})



# Deb-o-Matic - AutoPkgTest module
#
# Copyright (C) 2014 Leo Iannacone
# Copyright (C) 2015-2022 Luca Falavigna
#
# Authors: Leo Iannacone <l3on@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Run autpkgtest and stores output on top of the built
# package in the pool directory.
#
# Can be configured by adding this section to the config file:
#
# [autopkgtest]
# options: [adt-run options]
#

import os
from subprocess import call
from shutil import copyfile, rmtree


class DebomaticModule_AutoPkgTest:

    def __init__(self):
        self.adt = '/usr/bin/autopkgtest'
        self.options = []
        self.changesfile = None
        self.after = ['Lintian']

    def _set_up_testbed(self, args):
        """Performs:
         - check if dsc exists and declares a Testsuite
         - get autopkgtest options from the config
         - set up needed variables and files
        Returns True if test can be executed, False otherwise"""

        # check if dsc exists and it contains the Testsuite field
        autopkgtest_found = False
        for filename in os.listdir(self.resultdir):
            if filename.endswith('.dsc'):
                dsc = os.path.join(self.resultdir, filename)
                with open(dsc, 'r') as fd:
                    for line in fd:
                        if line.find('Testsuite: autopkgtest') >= 0:
                            autopkgtest_found = True
                            break
            if filename.endswith('_%s.changes' % args.architecture):
                self.changesfile = os.path.join(self.resultdir, filename)
        if not self.changesfile or not autopkgtest_found:
            return False

        # get autopkgtest options
        if args.opts.has_section('autopkgtest'):
            if args.opts.has_section('autopkgtest'):
                self.opts = args.opts.get('autopkgtest', 'options').split()

        # set up atd-run output dir
        self.resultdir_adt = os.path.join(self.resultdir, 'adt_out_dir')
        self.summary = 'log_summary'

        return True

    def post_build(self, args):
        if not args.success:
            return
        if not os.access(self.adt, os.X_OK):
            return
        if args.hostarchitecture:
            return

        self.resultdir = os.path.join(args.directory, 'pool', args.package)
        self.output = (os.path.join(self.resultdir, args.package) +
                       '.autopkgtest')
        if not self._set_up_testbed(args):
            return

        tempfiles = set()
        for pkgfile in args.files:
            target = os.path.join(self.resultdir, os.path.basename(pkgfile))
            if not os.path.exists(target):
                try:
                    copyfile(pkgfile, target)
                    tempfiles.add(target)
                except (IOError, OSError):
                    pass

        output = open(self.output, 'w')

        def write_header(header):
            output.write('┌{:─^70}┐\n'.format('─'))
            output.write('│ {: <69}│\n'.format(header))
            output.write('└{:─^70}┘\n'.format('─'))
            output.flush()

        def append_file(source, new_lines_at_the_end=True):
            if not source.startswith('/'):
                source = os.path.join(self.resultdir_adt, source)
            if os.path.isfile(source):
                with open(source, 'r') as fd:
                    for line in fd:
                        output.write(line)
                    if new_lines_at_the_end:
                        output.write('\n\n')
                output.flush()

        adt = [self.adt, '--apt-upgrade', '--output-dir', self.resultdir_adt,
               '--summary', os.path.join(self.resultdir_adt, self.summary),
               self.changesfile, '--', 'schroot',
               '%s-%s-debomatic' % (args.distribution, args.architecture)]
        if self.options:
            adt.insert(-4, self.options)

        # write the current adt-run command
        write_header('Command')
        output.write(' '.join(adt))
        output.write('\n\n\n')
        output.flush()

        # launch adt-run
        with open(os.devnull, 'w') as fd:
            call(adt, stdout=fd, stderr=fd)

        # build the log properly, first append the summary
        write_header('Tests summary')
        append_file(self.summary)

        # then the log
        write_header('Full log')
        append_file('log')

        # then the others
        all_files = [f for f in os.listdir(self.resultdir_adt) if
                     os.path.isfile(os.path.join(self.resultdir_adt, f)) and
                     f not in ['log', self.summary]]
        for curr_file in all_files:
            write_header('File: %s' % curr_file)
            append_file(curr_file, curr_file != all_files[-1])

        output.close()

        # clean up the system
        if (os.path.isdir(self.resultdir_adt)):
            rmtree(self.resultdir_adt)
        for pkgfile in tempfiles:
            os.remove(pkgfile)



# Deb-o-Matic - Lintian module
#
# Copyright (C) 2008-2009 David Futcher
# Copyright (C) 2008-2022 Luca Falavigna
#
# Authors: David Futcher <bobbo@ubuntu.com>
#          Luca Falavigna <dktrkranz@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Stores litian output on top of the built package in the pool directory.

import os
from shutil import copyfile
from subprocess import call, PIPE


class DebomaticModule_Lintian:

    def __init__(self):
        self.lintian = '/usr/bin/lintian'

    def post_build(self, args):
        if not args.success:
            return
        changesfile = None
        if args.opts.has_section('lintian'):
            lintopts = args.opts.get('lintian', 'options').strip()
        else:
            lintopts = []
        mirror = args.dists.get(args.distribution, 'mirror')
        try:
            profile = mirror.split('/')[-1]
        except IndexError:
            profile = None
        if profile:
            lintopts += ' --profile %s' % profile
        resultdir = os.path.join(args.directory, 'pool', args.package)
        lintian = os.path.join(resultdir, args.package) + '.lintian'
        files = os.listdir(resultdir)
        for filename in files:
            if filename.endswith('.changes'):
                changesfile = os.path.join(resultdir, filename)
                break
        tempfiles = set()
        for file in [f for f in args.files if not f.endswith('.changes') and
                     not f.endswith('.buildinfo')]:
            if not os.path.basename(file) in files:
                copyfile(file, os.path.join(resultdir, os.path.basename(file)))
                tempfiles.add(os.path.join(resultdir, os.path.basename(file)))
        if changesfile and os.access(self.lintian, os.X_OK):
            with open(lintian, 'w') as fd:
                call([self.lintian, '-V'], stdout=fd)
                fd.write('Options: %s\n\n' % lintopts)
                fd.flush()
                cmd = [self.lintian] + lintopts.split() + [changesfile]
                call(cmd, stdout=fd, stderr=PIPE)
        for file in tempfiles:
            os.remove(file)



# Deb-o-Matic - Blhc module
#
# Copyright (C) 2014 Mattia Rizzolo
# Copyright (C) 2015-2022 Luca Falavigna
#
# Authors: Mattia Rizzolo <mattia@mapreri.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option), any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Stores blhc output on top of the built package in the pool directory.

import os
from glob import glob
from subprocess import call
from tempfile import NamedTemporaryFile
from shutil import copyfile


class DebomaticModule_Blhc:

    def __init__(self):
        self.blhc = '/usr/bin/blhc'

    def post_build(self, args):
        if not args.success:
            return
        if args.opts.has_section('blhc') and \
           args.opts.has_option('blhc', 'options'):
            blhcopts = args.opts.get('blhc', 'options').strip().split()
        else:
            blhcopts = []
        if args.hostarchitecture:
            architecture = args.hostarchitecture
        else:
            architecture = args.architecture
        resultdir = os.path.join(args.directory, 'pool', args.package)
        buildlog = glob(os.path.join(resultdir, '*_%s.build' %
                                     architecture))[0]
        blhclog = os.path.join(resultdir, args.package) + '.blhc'
        if os.access(buildlog, os.R_OK):
            if os.access(self.blhc, os.X_OK):
                cmd = [self.blhc] + blhcopts + [buildlog]
                with NamedTemporaryFile() as fd:
                    exitcode = call(cmd, stdout=fd, stderr=fd)
                    # Save bhlc log only if useful information has been
                    # reported, discard output when exit status is:
                    # 0:   Buildlog is OK
                    # 1:   No compiler commands were found
                    if exitcode not in [0, 1]:
                        copyfile(fd.name, blhclog)
                        return
